package jarWithTestDependency.project2;

/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

import jarWithTestDependency.project1.Project1;

/**
 * 
 * 
 * @author <a href="mailto:markhobson@gmail.com">Mark Hobson</a>
 */
public class Project2
{
    // using a constant from JUnit requires junit dependency but is not detected since the constant
    // value is copied, then nothing can be detected in bytecode
    // Not the case any more, since we sniff it from the constant pool
    public final static int STATUS_ERROR = junit.runner.TestRunListener.STATUS_ERROR;

    public Project2()
    {
        Project1 project = new Project1();
    }
}
