/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.maven.shared.dependency.analyzer;

import java.util.Collection;
import java.util.Collections;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

/**
 * Patterns for classes
 */
public class ClassesPatterns {

    private final Collection<Pattern> patterns;

    /**
     * Default constructor.
     *
     * @param patterns a patterns to mach
     */
    public ClassesPatterns(Collection<String> patterns) {
        if (patterns == null) {
            this.patterns = Collections.emptyList();
        } else {
            this.patterns = patterns.stream().map(Pattern::compile).collect(Collectors.toSet());
        }
    }

    public ClassesPatterns() {
        this.patterns = Collections.emptySet();
    }

    public boolean isMatch(String className) {
        if (patterns.isEmpty()) {
            return false;
        }
        return patterns.stream().anyMatch(pattern -> pattern.matcher(className).matches());
    }
}
